                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                   compile                   */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipGrammar.h"

#include "SlipDictionary.h"
#include "SlipEnvironment.h"
#include "SlipEvaluate.h"
#include "SlipPool.h"
#include "SlipRead.h"
#include "SlipStack.h"

#include "SlipCompile.h"

/*---------------------------------- accepted forms ------------------------------------*/
/*                                                                                      */
/* <form>               ::= <essential form> | <non-essential form> | <quotation form>  */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* <essential form>     ::= <application> | <atom> | <define> | <if> | <lambda> |       */
/*                          <macro> | <set> | <symbol>                                  */
/*                                                                                      */
/* <non-essential form> ::= <and> | <begin> | <cond> | <cons-stream> | <delay> |        */
/*                          <eval> | <let> | <or> | <while>                             */
/*                                                                                      */
/* <quotation form>     ::= <quote> | <quasiquote> | <unquote> | <unquote-splicing>     */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* <and>                ::= (and <form>* )                                              */
/*                                                                                      */
/* <application>        ::= ( <form>+ )                                                 */
/*                                                                                      */
/* <atom>               ::= [atom]                                                      */
/*                                                                                      */
/* <begin>              ::= (begin <form>+ )                                            */
/*                                                                                      */
/* <cond>               ::= (cond <clause>+ ) | (cond <clause>* <else> )                */
/*                                                                                      */
/* <cons-stream>        ::= (cons-stream <form> <form> )                                */
/*                                                                                      */
/* <define>             ::= (define [symbol] <form> ) | (define <pattern> <form>+ )     */
/*                                                                                      */
/* <delay>              ::= (delay <form> )                                             */
/*                                                                                      */
/* <eval>               ::= (eval <form> )                                              */
/*                                                                                      */
/* <if>                 ::= (if <form> <form> ) | (if <form> <form> <form> )            */
/*                                                                                      */
/* <lambda>             ::= (lambda [<symbol>] <form>+ ) |                              */
/*                          (lambda ( <symbol>* ) <form>+ ) |                           */
/*                          (lambda ( <symbol>+ . <symbol> ) <form>+ )                  */
/*                                                                                      */
/* <let>                ::= (let ( <binding>* ) <form>+ )                               */
/*                                                                                      */
/* <macro>              ::= (macro [symbol] <form> ) | (macro <pattern> <form>+ )       */
/*                                                                                      */
/* <or>                 ::= (or <form>* )                                               */
/*                                                                                      */
/* <quote>              ::= (quote <form> )                                             */
/*                                                                                      */
/* <quasiquote>         ::= (quasiquote <form> )                                        */
/*                                                                                      */
/* <set>                ::= (set! [symbol] <form> )                                     */
/*                                                                                      */
/* <symbol>             ::= [symbol]                                                    */
/*                                                                                      */
/* <unquote>            ::= (unquote <form> )                                           */
/*                                                                                      */
/* <unquote-splicing>   ::= (unquote-splicing <form> )                                  */
/*                                                                                      */
/* <while>              ::= (while <form> <form>* )                                     */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* <binding>            ::= ( [symbol] <form> )                                         */
/*                                                                                      */
/* <clause>             ::= ( <form>+ )                                                 */
/*                                                                                      */
/* <else>               ::= (else <form>+ )                                             */
/*                                                                                      */
/* <pattern>            ::= ( [symbol]+ ) | ( [symbol]+ . [symbol] )                    */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/

/* Scheme compatibility

KEYWORDS
NO  =>
YES and → boolean predicate values required
YES begin
NO  case
YES cond → boolean predicate values required
EXT cons-stream
YES define → first class
NO  define-syntax
YES delay
NO  do
YES else
NEW eval
YES if → boolean predicate value required
YES lambda
NO  let
YES let* → let
NO  letrec
NO  let-syntax
NO  letrec-syntax
NEW load
NEW macro
NO  named let
YES or → boolean predicate values required
YES quasiquote
NEW quasisymbol
YES quote
YES set!
NO  syntax-rules
YES unquote
YES unquote-splicing

*/

/*---------------------------------- private constants ---------------------------------*/

static const TXT_type Compile_text = "compile";

/*---------------------------------- private prototypes --------------------------------*/

static EXP_type symbol_error(UNS_type,
                             SYM_type);

/*--------------------------------- protected variables --------------------------------*/

static EXP_type Body_Expression_P;
static EXP_type Car_Expression_P;
static EXP_type Cdr_Expression_P;
static DCT_type Dictionary_P;
static ENV_type Environment_P;
static EXP_type Expression_P;
static FRM_type Frame_P;
static IPT_type Input_Port_P;
static EXP_type List_Expression_P;
static MVL_type Macro_Value_P;
static PAI_type Pair_P;
static EXP_type Predicate_Expression_P;
static PRC_type Procedure_P;
static PRV_type Procedure_Vararg_P;
static SEQ_type Sequence_P;
static STR_type String_P;
static SYM_type Symbol_P;

/*---------------------------------- private functions ---------------------------------*/

static EXP_type error(UNS_type Error,
                      TXT_type Message_text)
  { Main_Error(Error,
               Message_text);
    return Unique_Undefined; }

static EXP_type keyword_error(UNS_type Error,
                              KEY_type Key)
  { SYM_type symbol;
    symbol = Pool_Keyword_Symbol(Key);
    return symbol_error(Error,
                        symbol); }

static EXP_type symbol_error(UNS_type Error,
                             SYM_type Symbol)
  { TXT_type message_text;
    message_text = Symbol->txt;
    return error(Error,
                 message_text); }

/*--------------------------------------------------------------------------------------*/

static UNS_type list(EXP_type List_expression,
                     KEY_type Key)
  { SGN_type size;
    size = Grammar_List_Size(List_expression);
    if (size < 0)
      keyword_error(ITF_error,
                    Key);
    return (UNS_type)size; }

static EXP_type single(EXP_type List_expression,
                       KEY_type Key)
  { EXP_type expression;
    PAI_type pair;
    UNS_type size;
    size = list(List_expression,
                Key);
    if (size > 1)
      return keyword_error(EX1_error,
                           Key);
    pair = (PAI_type)List_expression;
    expression = pair->car;
    return expression; }

/*--------------------------------------------------------------------------------------*/

static PAI_type pair_M(EXP_type Car_expression,
                       EXP_type Cdr_expression)
  { PAI_type pair;
    Car_Expression_P = Car_expression;
    Cdr_Expression_P = Cdr_expression;
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Cdr_Expression_P;
    return pair; }

static VOI_type unpair2(EXP_type Pair_Expression,
                        EXp_type Car_reference,
                        EXp_type Cdr_reference)
  { PAI_type pair;
    pair = Pair_Expression;
    *Car_reference = pair->car;
    if (Cdr_reference)
      *Cdr_reference = pair->cdr; }

static VOI_type unpair3(EXP_type Pair_Expression,
                        EXp_type Fst_reference,
                        EXp_type Scd_reference,
                        EXp_type Thd_reference)
  { PAI_type pair;
    pair = Pair_Expression;
    *Fst_reference = pair->car;
    pair           = pair->cdr;
    *Scd_reference = pair->car;
    if (Thd_reference)
      *Thd_reference = pair->cdr; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- expression-list ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static SEQ_type expression_list_M(EXP_type List_expression,
                                  KEY_type Key,
                                  BLN_type Boolean)
  { EXP_type expression,
             list_expression;
    SEQ_type sequence;
    UNS_type index,
             size;
    size = list(List_expression,
                Key);
    if (size == 0)
      return Unique_Empty_Sequence_P;
    Stack_Push_M(List_expression);
    sequence = make_SEQ_M(size);
    list_expression = Stack_Swap(sequence);
    for (index = 1;
         index < size;
         index += 1)
      { unpair2(list_expression,
                &Expression_P,
                &list_expression);
        Stack_Push_M(list_expression);
        expression = expression_M(Expression_P,
                                  FALSE);
        list_expression = Stack_Pop();
        sequence = Stack_Peek();
        sequence[index] = expression; }
    unpair2(list_expression,
            &expression,
            VOID);
    expression = expression_M(expression,
                              Boolean);
    sequence = Stack_Pop();
    sequence[size] = expression;
    return sequence; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- body ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type body_M(EXP_type List_expression,
                       KEY_type Key,
                       BLN_type Boolean)
  { EXP_type expression;
    UNS_type size;
    size = list(List_expression,
                Key);
    switch (size)
      { case 0:
          return keyword_error(AL1_error,
                               Key);
        case 1:
          unpair2(List_expression,
                  &expression,
                  VOID);
          return expression_M(expression,
                              Boolean);
        default:
          return expression_list_M(List_expression,
                                   Key,
                                   Boolean); }}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- inline --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type expression_M(EXP_type,
                             BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type inline_M(EXP_type Expression,
                         BLN_type Boolean)
  { BLN_type nested_eval;
    EXP_type expression;
    SMI_type frame_size_smallint;
    THK_type thunk;
    UNS_type frame_size;
    Stack_Push_M(Expression);
    Dictionary_Enter_Nested_Scope_M();
    expression = Stack_Peek();
    Expression_P = expression_M(expression,
                                TRUE);
    frame_size = Dictionary_Get_Frame_Size();
    nested_eval = Dictionary_Contains_Eval();
    Dictionary_P = Dictionary_Exit_Nested_Scope();
    if ((frame_size > 0) || nested_eval)
      { Stack_Zap();
        frame_size_smallint = make_SMU(frame_size);
        thunk = make_THK_M();
        thunk->exp = Expression_P;
        thunk->siz = frame_size_smallint;
        thunk->dct = Dictionary_P;
        if (Boolean)
          tail_THK(thunk);
        return thunk; }
    expression = Stack_Pop();
    return expression_M(expression,
                        Boolean); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- inline body ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type body_M(EXP_type,
                       KEY_type,
                       BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type inline_body_M(EXP_type List_expression,
                              KEY_type Key,
                              BLN_type Boolean)
  { BLN_type nested_eval;
    DCT_type dictionary;
    EXP_type list_expression;
    SMI_type frame_size_smallint;
    THK_type thunk;
    UNS_type frame_size;
    Stack_Push_M(List_expression);
    Dictionary_Enter_Nested_Scope_M();
    list_expression = Stack_Peek();
    Expression_P = body_M(list_expression,
                          Key,
                          TRUE);
    frame_size = Dictionary_Get_Frame_Size();
    nested_eval = Dictionary_Contains_Eval();
    dictionary = Dictionary_Exit_Nested_Scope();
    if ((frame_size > 0) || nested_eval)
      { Stack_Zap();
        frame_size_smallint = make_SMU(frame_size);
        thunk = make_THK_M();
        thunk->exp = Expression_P;
        thunk->siz = frame_size_smallint;
        thunk->dct = dictionary;
        if (Boolean)
          tail_THK(thunk);
        return thunk; }
    list_expression = Stack_Pop();
    return body_M(list_expression,
                  Key,
                  Boolean); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- inline-list ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type inline_M(EXP_type,
                         BLN_type);
static UNS_type     list(EXP_type,
                         KEY_type);
static VOI_type  unpair2(EXP_type,
                         EXp_type,
                         EXp_type);

/*--------------------------------------------------------------------------------------*/

static SEQ_type inline_list_M(EXP_type List_expression,
                              KEY_type Key,
                              BLN_type Boolean)
  { EXP_type expression,
             list_expression;
    SEQ_type sequence;
    UNS_type index,
             size;
    size = list(List_expression,
                Key);
    if (size == 0)
      return Unique_Empty_Sequence_P;
    Stack_Push_M(List_expression);
    sequence = make_SEQ_M(size);
    list_expression = Stack_Swap(sequence);
    for (index = 1;
         index < size;
         index += 1)
      { unpair2(list_expression,
                &Expression_P,
                &list_expression);
        Stack_Push_M(list_expression);
        expression = inline_M(Expression_P,
                              FALSE);
        list_expression = Stack_Pop();
        sequence = Stack_Peek();
        sequence[index] = expression; }
    unpair2(list_expression,
            &expression,
            VOID);
    expression = inline_M(expression,
                          Boolean);
    sequence = Stack_Pop();
    sequence[size] = expression;
    return sequence; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- macro value ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------------*/

static EXP_type macro_value_M(MVL_type Macro_value)
  { DCT_type dictionary;
    ENV_type environment;
    EXP_type expression;
    FRM_type frame;
    Macro_Value_P = Macro_value;
    environment = Environment_Get_Current_Environment();
    Stack_Push_M(environment);
    frame = Environment_Get_Current_Frame();
    Stack_Push_M(frame);
    dictionary = Dictionary_Get_Current_Dictionary();
    Stack_Push_M(dictionary);
    expression  = Macro_Value_P->exp;
    environment = Macro_Value_P->env;
    frame       = Macro_Value_P->frm;
    dictionary  = Macro_Value_P->dct;
    Environment_Replace_Environment(environment);
    Environment_Replace_Frame(frame);
    Dictionary_Replace_Dictionary(dictionary);
    expression = Evaluate_M(expression);
    dictionary = Stack_Pop();
    frame = Stack_Pop();
    environment = Stack_Pop();
    Environment_Replace_Environment(environment);
    Environment_Replace_Frame(frame);
    Dictionary_Replace_Dictionary(dictionary);
    expression = expression_M(expression,
                              FALSE);                                               // ???
    return expression; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- parameters -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type  unpair2(EXP_type,
                         EXp_type,
                         EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type parameters_M(EXP_type List_expression,
                             KEY_type Key)
  { EXP_type symbol_expression;
    UNS_type offset;
    for (List_Expression_P = List_expression;
         is_PAI(List_Expression_P); )
      { unpair2(List_Expression_P,
                &symbol_expression,
                &List_Expression_P);
        if (!is_SYM(symbol_expression))
          return keyword_error(IPA_error,
                               Key);
        offset = Dictionary_Define_Variable_M(symbol_expression);
        if (offset == 0)
          symbol_error(IVS_error,
                       symbol_expression); }
    return List_Expression_P; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- and ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static SEQ_type inline_list_M(EXP_type,
                              KEY_type,
                              BLN_type);

/*--------------------------------------------------------------------------------------*/

static AND_type and_M(BLN_type Boolean)
  { AND_type and;
    and = make_AND_M();
    and->prd = inline_list_M(List_Expression_P,
                             Pool_And,
                             Boolean);
    return and; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- application ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type      expression_M(EXP_type,
                                  BLN_type);
static SEQ_type expression_list_M(EXP_type,
                                  KEY_type,
                                  BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type application_M(BLN_type Boolean)
  { APL_type application;
    EXP_type expression,
             list_expression;
    GLA_type global_application;
    LCA_type local_application;
    TAG_type tag;
    if (is_NUL(List_Expression_P))
      { Expression_P = expression_M(Expression_P,
                                    Boolean);
        Sequence_P = Unique_Empty_Sequence_P; }
    else
      { Stack_Push_M(List_Expression_P);
        expression = expression_M(Expression_P,
                                  FALSE);
        list_expression = Stack_Swap(expression);
        Sequence_P = expression_list_M(list_expression,
                                       Pool_Application,
                                       Boolean);
        Expression_P = Stack_Pop(); }
    tag = Grammar_Tag(Expression_P);
    switch (tag)
      { case GLB_tag:
          global_application = make_GLA_M();
          global_application->glb = Expression_P;
          global_application->arg = Sequence_P;
          if (Boolean)
            tail_GLA(global_application);
          return global_application;
        case LCL_tag:
          local_application = make_LCA_M();
          local_application->lcl = Expression_P;
          local_application->arg = Sequence_P;
          if (Boolean)
            tail_LCA(local_application);
          return local_application;
        default:
          application = make_APL_M();
          application->exp = Expression_P;
          application->arg = Sequence_P;
          if (Boolean)
            tail_APL(application);
          return application; }}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- begin ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type body_M(EXP_type,
                       KEY_type,
                       BLN_type);

/*--------------------------------------------------------------------------------------*/

static BEG_type begin_M(BLN_type Boolean)
  { BEG_type begin;
    begin = make_BEG_M();
    begin->exp = body_M(List_Expression_P,
                        Pool_Begin,
                        Boolean);
    return begin; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- cond ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type        body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type      inline_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type cond_M(BLN_type Boolean)
  { BLN_type no_else_clause;
    CND_type cond;
    CNE_type cond_else;
    EXP_type body_expression,
             clause_expression,
             list_expression,
             predicate_expression;
    SEQ_type body_sequence,
             predicate_sequence;
    UNS_type index,
             size;
    no_else_clause = TRUE;
    list_expression = List_Expression_P;
    for (size = 0;
         is_PAI(list_expression);
         size += 1)
      { unpair2(list_expression,
                &clause_expression,
                &list_expression);
        unpair2(clause_expression,
                &predicate_expression,
                &Body_Expression_P);
        if (is_SYM(predicate_expression))
          if (predicate_expression == Pool_Keyword_Symbol(Pool_Else))
            { no_else_clause = FALSE;
              break; }}
    if (!is_NUL(list_expression))
      return keyword_error(ITF_error,
                           Pool_Cond);
    if (size == 0)
      { if (no_else_clause)
          return keyword_error(AL1_error,
                               Pool_Cond);
        Body_Expression_P = inline_body_M(Body_Expression_P,
                                          Pool_Cond,
                                          Boolean);
        cond_else = make_CNE_M();
        cond_else->prd = Unique_Empty_Sequence_P;
        cond_else->bod = Unique_Empty_Sequence_P;
        cond_else->els = Body_Expression_P;
        return cond_else; }
    predicate_sequence = make_SEQ_M(size);
    Stack_Push_M(predicate_sequence);
    body_sequence = make_SEQ_M(size);
    Stack_Push_M(body_sequence);
    for (index = 1;
         index <= size;
         index += 1)
      { unpair2(List_Expression_P,
                &clause_expression,
                &List_Expression_P);
        unpair2(clause_expression,
                &Predicate_Expression_P,
                &Body_Expression_P);
        Stack_Push_M(List_Expression_P);
        Stack_Push_M(Body_Expression_P);
        predicate_expression = inline_M(Predicate_Expression_P,
                                        FALSE);
        body_expression = Stack_Swap(predicate_expression);
        body_expression = inline_body_M(body_expression,
                                        Pool_Cond,
                                        Boolean);
        predicate_expression = Stack_Pop();
        List_Expression_P = Stack_Pop();
        body_sequence = Stack_Pop();
        body_sequence[index] = body_expression;
        predicate_sequence = Stack_Peek();
        predicate_sequence[index] = predicate_expression;
        Stack_Push_M(body_sequence); }
    if (no_else_clause)
      { cond = make_CND_M();
        cond->bod = Stack_Pop();
        cond->prd = Stack_Pop();
        return cond; }
    unpair2(List_Expression_P,
            &clause_expression,
            VOID);
    unpair2(clause_expression,
            &predicate_expression,
            &body_expression);
    Body_Expression_P = inline_body_M(body_expression,
                                      Pool_Cond,
                                      Boolean);
    cond_else = make_CNE_M();
    cond_else->bod = Stack_Pop();
    cond_else->prd = Stack_Pop();
    cond_else->els = Body_Expression_P;
    return cond_else; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- cons-stream ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair3(EXP_type,
                              EXp_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static CNS_type cons_stream_M(VOI_type)
  { CNS_type cons_stream;
    DCT_type dictionary;
    EXP_type delayed_expression;
    SMI_type size_smallint;
    UNS_type frame_size,
             size;
    size = list(List_Expression_P,
                Pool_Cons_Stream);
    if (size != 2)
      return keyword_error(EX2_error,
                           Pool_Cons_Stream);
    unpair3(List_Expression_P,
            &Expression_P,
            &delayed_expression,
            VOID);
    Stack_Push_M(delayed_expression);
    Expression_P = expression_M(Expression_P,
                                FALSE);
    Dictionary_Enter_Nested_Scope_M();
    delayed_expression = Stack_Swap(Expression_P);

    Expression_P = expression_M(delayed_expression,
                                TRUE);
    frame_size = Dictionary_Get_Frame_Size();
    cons_stream = make_CNS_M();
    dictionary = Dictionary_Exit_Nested_Scope();
    size_smallint = make_SMU(frame_size);
    cons_stream->car = Stack_Pop();
    cons_stream->cdr = Expression_P;
    cons_stream->siz = size_smallint;
    cons_stream->dct = dictionary;
    return cons_stream; }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- define ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type        body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static EXP_type  parameters_M(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type define_M(VOI_type)
  { DCT_type dictionary;
    DFF_type define_function;
    DFV_type define_variable;
    DFZ_type define_function_vararg;
    EXP_type expression,
             pattern_expression,
             residue_expression,
             symbol_expression;
    SMI_type frame_size_smallint,
             offset_smallint,
             parameter_count_smallint;
    SYM_type symbol;
    TAG_type tag;
    UNS_type frame_size,
             size,
             offset,
             parameter_count;
    size = list(List_Expression_P,
                Pool_Define);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_Define);
    unpair2(List_Expression_P,
            &pattern_expression,
            &Expression_P);
    tag = Grammar_Tag(pattern_expression);
    switch (tag)
      { case SYM_tag:
          if (size != 2)
            return keyword_error(EX2_error,
                                 Pool_Define);
          Stack_Push_M(pattern_expression);
          unpair2(Expression_P,
                  &expression,
                  VOID);
          Expression_P = expression_M(expression,
                                      FALSE);
          symbol = Stack_Pop();
          offset = Dictionary_Define_Variable_M(symbol);
          if (offset == 0)
            symbol_error(IVS_error,
                         symbol);
          offset_smallint = make_SMU(offset);
          define_variable = make_DFV_M();
          define_variable->exp = Expression_P;
          define_variable->ofs = offset_smallint;
          return define_variable;
        case PAI_tag:
          unpair2(pattern_expression,
                  &symbol_expression,
                  &List_Expression_P);
          if (!is_SYM(symbol_expression))
            return keyword_error(IVV_error,
                                 Pool_Define);
          Stack_Push_M(symbol_expression);
          symbol = Stack_Peek();
          offset = Dictionary_Define_Variable_M(symbol);
          if (offset == 0)
            symbol_error(IVS_error,
                         symbol);
          offset_smallint = make_SMU(offset);
          Dictionary_Enter_Nested_Scope_M();
          residue_expression = parameters_M(List_Expression_P,
                                            Pool_Define);
          parameter_count = Dictionary_Get_Frame_Size();
          parameter_count_smallint = make_SMU(parameter_count);
          tag = Grammar_Tag(residue_expression);
          switch (tag)
            { case NUL_tag:
                Expression_P = body_M(Expression_P,
                                      Pool_Define,
                                      TRUE);
                frame_size = Dictionary_Get_Frame_Size();
                frame_size_smallint = make_SMU(frame_size);
                define_function = make_DFF_M();
                dictionary = Dictionary_Exit_Nested_Scope();
                define_function->sym = Stack_Pop();
                define_function->ofs = offset_smallint;
                define_function->par = parameter_count_smallint;
                define_function->exp = Expression_P;
                define_function->siz = frame_size_smallint;
                define_function->dct = dictionary;
                return define_function;
              case SYM_tag:
                Dictionary_Define_Variable_M(residue_expression);
                Expression_P = body_M(Expression_P,
                                      Pool_Define,
                                      TRUE);
                frame_size = Dictionary_Get_Frame_Size();
                frame_size_smallint = make_SMU(frame_size);
                define_function_vararg = make_DFZ_M();
                dictionary = Dictionary_Exit_Nested_Scope();
                define_function_vararg->sym = Stack_Pop();
                define_function_vararg->ofs = offset_smallint;
                define_function_vararg->par = parameter_count_smallint;
                define_function_vararg->exp = Expression_P;
                define_function_vararg->siz = frame_size_smallint;
                define_function_vararg->dct = dictionary;
                return define_function_vararg;
              default:
                return keyword_error(IPA_error,
                                     Pool_Define); }
        default:
          return keyword_error(IPA_error,
                               Pool_Define); }}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- delay ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type expression_M(EXP_type,
                             BLN_type);
static EXP_type       single(EXP_type,
                             BLN_type);

/*--------------------------------------------------------------------------------------*/

static DEL_type delay_M(VOI_type)
  { DCT_type dictionary;
    DEL_type delay;
    SMI_type frame_size_smallint;
    UNS_type frame_size;
    Expression_P = single(List_Expression_P,
                          Pool_Delay);
    Dictionary_Enter_Nested_Scope_M();
    Expression_P = expression_M(Expression_P,
                                  TRUE);
    frame_size = Dictionary_Get_Frame_Size();
    frame_size_smallint = make_SMU(frame_size);
    delay = make_DEL_M();
    dictionary = Dictionary_Exit_Nested_Scope();
    delay->exp = Expression_P;
    delay->siz = frame_size_smallint;
    delay->dct = dictionary;
    return delay; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- eval ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type expression_M(EXP_type,
                             BLN_type);
static EXP_type       single(EXP_type,
                             BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type eval_M(BLN_type Boolean)
  { EVA_type eval;
    EXP_type expression;
    expression = single(List_Expression_P,
                        Pool_Eval);
    Dictionary_Register_Eval();
    Expression_P = expression_M(expression,
                                Boolean);
    eval = make_EVA_M();
    eval->exp = Expression_P;
    return eval; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- if -----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type      inline_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);
static VOI_type       unpair3(EXP_type,
                              EXp_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type if_M(BLN_type Boolean)
  { EXP_type alternate_expression,
             consequent_expression,
             list_expression,
             predicate_expression;
    IFD_type if_double;
    IFS_type if_single;
    UNS_type size;
    size = list(List_Expression_P,
                Pool_If);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_If);
    if (size > 3)
      return keyword_error(TMX_error,
                           Pool_If);
    unpair2(List_Expression_P,
            &Expression_P,
            &list_expression);
    Stack_Push_M(list_expression);
    predicate_expression = expression_M(Expression_P,
                                        FALSE);
    list_expression = Stack_Swap(predicate_expression);
    if (size == 2)
      { unpair2(list_expression,
                &consequent_expression,
                VOID);
        Expression_P = inline_M(consequent_expression,
                                Boolean);
        if_single = make_IFS_M();
        if_single->cns = Expression_P;
        if_single->prd = Stack_Pop();
        return if_single; }
    unpair3(list_expression,
            &Expression_P,
            &alternate_expression,
            VOID);
    Stack_Push_M(alternate_expression);
    consequent_expression = inline_M(Expression_P,
                                     Boolean);
    alternate_expression = Stack_Swap(consequent_expression);
    Expression_P = inline_M(alternate_expression,
                            Boolean);
    if_double = make_IFD_M();
    if_double->alt = Expression_P;
    if_double->cns = Stack_Pop();
    if_double->prd = Stack_Pop();
    return if_double; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- lambda --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type        body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static EXP_type  parameters_M(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type lambda_M(VOI_type)
  { DCT_type dictionary;
    EXP_type residue_expression;
    LMB_type lambda;
    LMV_type lambda_vararg;
    SMI_type frame_size_smallint,
             parameter_count_smallint;
    TAG_type tag;
    UNS_type offset,
             frame_size,
             parameter_count,
             size;
    size = list(List_Expression_P,
                Pool_Lambda);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_Lambda);
    Dictionary_Enter_Nested_Scope_M();
    unpair2(List_Expression_P,
            &List_Expression_P,
            &Expression_P);
    residue_expression = parameters_M(List_Expression_P,
                                      Pool_Lambda);
    parameter_count = Dictionary_Get_Frame_Size();
    parameter_count_smallint = make_SMU(parameter_count);
    tag = Grammar_Tag(residue_expression);
    switch (tag)
      { case NUL_tag:
          Expression_P = body_M(Expression_P,
                                Pool_Lambda,
                                TRUE);
          frame_size = Dictionary_Get_Frame_Size();
          frame_size_smallint = make_SMU(frame_size);
          lambda = make_LMB_M();
          dictionary = Dictionary_Exit_Nested_Scope();
          lambda->par = parameter_count_smallint;
          lambda->exp = Expression_P;
          lambda->siz = frame_size_smallint;
          lambda->dct = dictionary;
          return lambda;
        case SYM_tag:
          offset = Dictionary_Define_Variable_M(residue_expression);
          if (offset == 0)
            symbol_error(IVS_error,
                         residue_expression);
          Expression_P = body_M(Expression_P,
                                Pool_Lambda,
                                TRUE);
          frame_size = Dictionary_Get_Frame_Size();
          frame_size_smallint = make_SMU(frame_size);
          lambda_vararg = make_LMV_M();
          dictionary = Dictionary_Exit_Nested_Scope();
          lambda_vararg->par = parameter_count_smallint;
          lambda_vararg->exp = Expression_P;
          lambda_vararg->siz = frame_size_smallint;
          lambda_vararg->dct = dictionary;
          return lambda_vararg;
          break;
        default:
          return keyword_error(IPA_error,
                               Pool_Lambda); }}

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- let ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type        body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);
static VOI_type       unpair3(EXP_type,
                              EXp_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static LET_type let_M(BLN_type Boolean)
  { DCT_type dictionary;
    EXP_type expression,
             list_expression,
             symbol_expression;
    LET_type let_P;
    SEQ_type sequence;
    SMI_type frame_size_smallint;
    UNS_type binding_size,
             frame_size,
             index,
             offset,
             size;
    size = list(List_Expression_P,
                Pool_Let);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_Let);
    unpair2(List_Expression_P,
            &List_Expression_P,
            &expression);
    Stack_Push_M(expression);
    Dictionary_Enter_Nested_Scope_M();
    size = list(List_Expression_P,
                Pool_Let);
    if (size == 0)
      sequence = Unique_Empty_Sequence_P;
    else
      { sequence = make_SEQ_M(size);
        Stack_Push_M(sequence);
        for (index = 1;
             index <= size;
             index += 1)
          { unpair2(List_Expression_P,
                    &List_Expression_P,
                    &list_expression);
            Stack_Push_M(list_expression);
            binding_size = list(List_Expression_P,
                                Pool_Let);
            if (binding_size != 2)
              return keyword_error(BND_error,
                                   Pool_Let);
            unpair3(List_Expression_P,
                    &symbol_expression,
                    &Expression_P,
                    VOID);
            if (!is_SYM(symbol_expression))
              return keyword_error(IVV_error,
                                   Pool_Let);
            Stack_Push_M(symbol_expression);
            expression = inline_M(Expression_P,
                                  FALSE);
            symbol_expression = Stack_Pop();
            List_Expression_P = Stack_Pop();
            sequence = Stack_Peek();
            sequence[index] = expression;
            offset = Dictionary_Define_Variable_M(symbol_expression);
            if (offset == 0)
              return symbol_error(IVS_error,
                                  symbol_expression); }
        sequence = Stack_Pop(); }
    list_expression = Stack_Swap(sequence);
    Expression_P = body_M(list_expression,
                          Pool_Let,
                          TRUE);
    frame_size = Dictionary_Get_Frame_Size();
    frame_size_smallint = make_SMU(frame_size);
    let_P = make_LET_M();
    dictionary = Dictionary_Exit_Nested_Scope();
    let_P->seq = Stack_Pop();
    let_P->exp = Expression_P;
    let_P->siz = frame_size_smallint;
    let_P->dct = dictionary;
    if (Boolean)
      tail_LET(let_P);
    return let_P; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- load ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static EXP_type        single(EXP_type,
                              BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type load_M(BLN_type Boolean)
  { ADR_type address;
    EXP_type expression,
             string_expression;
    TXT_type text;
    UNS_type length;
    string_expression = single(List_Expression_P,
                               Pool_Load);
    if (!is_STR(string_expression))
      return keyword_error(NAS_error,
                           Pool_Load);
    String_P = string_expression;
    text = String_P->txt;
    address = Main_Open_Input_Port_M(text);
    if (!address)
      return keyword_error(OIF_error,
                           Pool_Load);
    length = length_STR(String_P);
    Input_Port_P = make_IPT_M(length);
    text = String_P->txt;
    Input_Port_P->adr = address;
    C_STRCPY(Input_Port_P->txt, text);
    expression = Read_Port_M(Input_Port_P);
    if (!Main_Close_Input_Port(address))
      return keyword_error(CIP_error,
                           Pool_Load);
    Input_Port_P->adr = VOID;
    return expression_M(expression,
                        Boolean); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- macro ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type        body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static EXP_type  parameters_M(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static MAC_type macro_M(VOI_type)
  { DCT_type dictionary;
    EXP_type pattern_expression,
             residue_expression,
             symbol_expression;
    FRM_type frame;
    MAC_type macro;
    SMI_type frame_size_smallint,
             parameter_count_smallint;
    SYM_type symbol;
    TAG_type tag;
    UNS_type offset,
             frame_size,
             size,
             parameter_count;
    size = list(List_Expression_P,
                Pool_Macro);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_Macro);
    unpair2(List_Expression_P,
            &pattern_expression,
            &Body_Expression_P);
    tag = Grammar_Tag(pattern_expression);
    switch (tag)
      { case SYM_tag:
          if (size != 2)
            return keyword_error(EX2_error,
                                 Pool_Macro);
          Stack_Push_M(pattern_expression);
          unpair2(Body_Expression_P,
                  &Expression_P,
                  VOID);
          dictionary = Dictionary_Activate_Macro();
          Stack_Push_M(dictionary);
          frame = Environment_Activate_Macro();
          Stack_Push_M(frame);
          Expression_P = expression_M(Expression_P,
                                      FALSE);
          frame_size = Dictionary_Get_Frame_Size();
          Environment_Adjust_Current_Frame_Size_M(frame_size);
          Macro_Value_P = make_MVL_M();
          Macro_Value_P->exp = Expression_P;
          Macro_Value_P->env = Environment_Get_Current_Environment();
          Macro_Value_P->frm = Environment_Get_Current_Frame();
          Macro_Value_P->dct = Dictionary_Get_Current_Dictionary();
          macro = make_MAC_M();
          frame = Stack_Pop();
          dictionary = Stack_Pop();
          symbol = Stack_Pop();
          macro->sym = symbol;
          Dictionary_Deactivate_Macro(dictionary);
          Environment_Deactivate_Macro(frame);
          if (Dictionary_Define_Macro_Symbol_M(symbol,
                                               Macro_Value_P))
            return symbol_error(IMS_error,
                                symbol);
          return macro;
        case PAI_tag:
          unpair2(pattern_expression,
                  &symbol_expression,
                  &List_Expression_P);
          if (!is_SYM(symbol_expression))
            return keyword_error(IVV_error,
                                 Pool_Macro);
          Stack_Push_M(symbol_expression);
          dictionary = Dictionary_Activate_Macro();
          Stack_Push_M(dictionary);
          Dictionary_Enter_Nested_Scope_M();
          residue_expression = parameters_M(List_Expression_P,
                                            Pool_Macro);
          parameter_count = Dictionary_Get_Frame_Size();
          parameter_count_smallint = make_SMU(parameter_count);
          tag = Grammar_Tag(residue_expression);
          switch (tag)
            { case NUL_tag:
                Body_Expression_P = body_M(Body_Expression_P,
                                           Pool_Macro,
                                           TRUE);
                Frame_P = Environment_Activate_Macro();
                Environment_P = Environment_Extend_M();
                Environment_Deactivate_Macro(Frame_P);
                frame_size = Dictionary_Get_Frame_Size();
                frame_size_smallint = make_SMU(frame_size);
                Dictionary_P = Dictionary_Exit_Nested_Scope();
                Procedure_P = make_PRC_M();
                dictionary = Stack_Pop();
                symbol = Stack_Pop();
                Procedure_P->sym = symbol;
                Procedure_P->par = parameter_count_smallint;
                Procedure_P->exp = Body_Expression_P;
                Procedure_P->siz = frame_size_smallint;
                Procedure_P->env = Environment_P;
                Procedure_P->dct = Dictionary_P;
                macro = make_MAC_M();
                macro->sym = symbol;
                Dictionary_Deactivate_Macro(dictionary);
                if (Dictionary_Define_Macro_Symbol_M(symbol,
                                                     Procedure_P))
                  return symbol_error(IMS_error,
                                      symbol);

                return macro;
              case SYM_tag:
                offset = Dictionary_Define_Variable_M(residue_expression);
                if (offset == 0)
                  symbol_error(IVS_error,
                               residue_expression);
                Body_Expression_P = body_M(Body_Expression_P,
                                           Pool_Macro,
                                           TRUE);
                Frame_P = Environment_Activate_Macro();
                Environment_P = Environment_Extend_M();
                Environment_Deactivate_Macro(Frame_P);
                frame_size = Dictionary_Get_Frame_Size();
                frame_size_smallint = make_SMU(frame_size);
                Dictionary_P = Dictionary_Exit_Nested_Scope();
                Procedure_Vararg_P = make_PRV_M();
                dictionary = Stack_Pop();
                symbol = Stack_Pop();
                Procedure_Vararg_P->sym = symbol;
                Procedure_Vararg_P->par = parameter_count_smallint;
                Procedure_Vararg_P->exp = Body_Expression_P;
                Procedure_Vararg_P->siz = frame_size_smallint;
                Procedure_Vararg_P->env = Environment_P;
                Procedure_Vararg_P->dct = Dictionary_P;
                macro = make_MAC_M();
                macro->sym = symbol;
                Dictionary_Deactivate_Macro(dictionary);
                if (Dictionary_Define_Macro_Symbol_M(symbol,
                                                     Procedure_Vararg_P))
                  return symbol_error(IMS_error,
                                      symbol);
                return macro;
              default:
                return keyword_error(IPA_error,
                                     Pool_Macro); }
        default:
          return keyword_error(IPA_error,
                               Pool_Macro); }}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------- macro application ---------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type expression_M(EXP_type,
                             BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type macro_application_M(EXP_type Macro_expression,
                                    BLN_type Boolean)
  { EXP_type expression;
    expression = Evaluate_Macro_M(Macro_expression,
                                  List_Expression_P);
    return expression_M(expression,
                        Boolean); }                                                 // ???

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- or -----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static SEQ_type inline_list_M(EXP_type,
                              KEY_type,
                              BLN_type);

/*--------------------------------------------------------------------------------------*/

static ORR_type or_M(EXP_type List_expression,
                     BLN_type Boolean)
  { ORR_type or;
    or = make_ORR_M();
    or->prd =  inline_list_M(List_expression,
                             Pool_Or,
                             Boolean);
    return or; }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- quasiquote ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

typedef
  enum { No_splicing,
         List_splicing,
         Vector_splicing } SPL_type;

/*--------------------------------------------------------------------------------------*/

static EXP_type      keyword_error(UNS_type,
                                   KEY_type);
static EXP_type quasi_expression_M(EXP_type,
                                   SPL_type,
                                   UNS_type,
                                   UNS_type);
static EXP_type             single(EXP_type,
                                   BLN_type);
static VOI_type            unpair2(EXP_type,
                                   EXp_type,
                                   EXp_type);

/*--------------------------------------------------------------------------------------*/

static PAI_type quasi_pair_M(PAI_type Pair,
                             UNS_type Counter,
                             UNS_type Nesting_level)
  { EXP_type car_expression,
             cdr_expression;
    PAI_type pair;
    unpair2(Pair,
            &Car_Expression_P,
            &cdr_expression);
    Stack_Push_M(cdr_expression);
    car_expression = quasi_expression_M(Car_Expression_P,
                                        List_splicing,
                                        Counter,
                                        Nesting_level);
    cdr_expression = Stack_Swap(car_expression);
    cdr_expression = quasi_expression_M(cdr_expression,
                                        List_splicing,
                                        Counter,
                                        Nesting_level);
    car_expression = Stack_Pop();
    pair = pair_M(car_expression,
                  cdr_expression);
    return pair; }

/*-----------------------------------quasi-quasiquote ----------------------------------*/

static EXP_type quasi_quasiquote_M(EXP_type List_expression,
                                   UNS_type Counter,
                                   UNS_type Nesting_level)
  { EXP_type expression;
    PAI_type quoted_pair;
    QQU_type quasiquote;
    expression = single(List_expression,
                        Pool_Quasiquote);
    Expression_P = quasi_expression_M(expression,
                                      No_splicing,
                                      Counter,
                                      Nesting_level + 1);
    if (Nesting_level % 2 == 1)
      { quoted_pair = Pool_Quotation_M(Pool_Quasiquote,
                                       Expression_P);
        return quoted_pair; }
    quasiquote = make_QQU_M();
    quasiquote->exp = Expression_P;
    return quasiquote; }

/*-----------------------------------quasi-quasisymbol ---------------------------------*/

static EXP_type quasi_quasisymbol_M(EXP_type List_expression,
                                    UNS_type Counter,
                                    UNS_type Nesting_level)
  { QSY_type quasisymbol;
    PAI_type quoted_pair;
    SMI_type counter_smallint;
    Expression_P = single(List_expression,
                          Pool_Quasisymbol);
    if (!is_SYM(Expression_P))
       return keyword_error(NSY_error,
                            Pool_Quasisymbol);
    if (Nesting_level % 2 == 0)
      { quoted_pair = Pool_Quotation_M(Pool_Quasisymbol,
                                       Expression_P);
        return quoted_pair; }
    counter_smallint = make_SMU(Counter);
    quasisymbol = make_QSY_M();
    quasisymbol->sym = Expression_P;
    quasisymbol->ctr = counter_smallint;
    return quasisymbol; }

/*------------------------------------quasi-unquote ------------------------------------*/

static EXP_type quasi_unquote_M(EXP_type List_expression,
                                UNS_type Counter,
                                UNS_type Nesting_level)
  { EXP_type expression;
    PAI_type quoted_pair;
    UNQ_type unquote;
    expression = single(List_expression,
                        Pool_Unquote);
    Expression_P = quasi_expression_M(expression,
                                      No_splicing,
                                      Counter,
                                      Nesting_level + 1);
    if (Nesting_level % 2 == 0)
      { quoted_pair = Pool_Quotation_M(Pool_Unquote,
                                       Expression_P);
        return quoted_pair; }
    unquote = make_UNQ_M();
    unquote->exp = Expression_P;
    return unquote; }


/*-----------------------------quasi-unquote-list-splicing -----------------------------*/

static EXP_type quasi_unquote_list_splicing_M(EXP_type List_expression,
                                              UNS_type Counter,
                                              UNS_type Nesting_level)
  { EXP_type expression;
    PAI_type quoted_pair;
    ULS_type unquote_list_splice;
    expression = single(List_expression,
                        Pool_Unquote_Splicing);
    Expression_P = quasi_expression_M(expression,
                                      No_splicing,
                                      Counter,
                                      Nesting_level + 1);
    if (Nesting_level % 2 == 0)
      { quoted_pair = Pool_Quotation_M(Pool_Unquote_Splicing,
                                       Expression_P);
        return quoted_pair; }
    unquote_list_splice = make_ULS_M();
    unquote_list_splice->exp = Expression_P;
    return unquote_list_splice; }

/*----------------------------quasi-unquote-vector-splicing ----------------------------*/

static EXP_type quasi_unquote_vector_splicing_M(EXP_type List_expression,
                                                UNS_type Counter,
                                                UNS_type Nesting_level)
  { EXP_type expression;
    PAI_type quoted_pair;
    UVS_type unquote_vector_splice;
    expression = single(List_expression,
                        Pool_Unquote_Splicing);
    Expression_P = quasi_expression_M(expression,
                                      No_splicing,
                                      Counter,
                                      Nesting_level + 1);
    if (Nesting_level % 2 == 0)
      { quoted_pair = Pool_Quotation_M(Pool_Unquote_Splicing,
                                       Expression_P);
        return quoted_pair; }
    unquote_vector_splice = make_UVS_M();
    unquote_vector_splice->exp = Expression_P;
    return unquote_vector_splice; }

/*------------------------------------quasi-literal ------------------------------------*/

static EXP_type quasi_literal(EXP_type Literal_expression)
  { return Literal_expression; }

/*-------------------------------------quasi-form --------------------------------------*/

static PAI_type quasi_form_M(PAI_type Pair,
                             SPL_type Splicing_status,
                             UNS_type Counter,
                             UNS_type Nesting_level)
  { EXP_type car_expression,
             list_expression;
    unpair2(Pair,
            &car_expression,
            &list_expression);
    if (car_expression == Pool_Keyword_Symbol(Pool_Quasiquote))
      return quasi_quasiquote_M(list_expression,
                                Counter,
                                Nesting_level);
    if (car_expression == Pool_Keyword_Symbol(Pool_Unquote))
      return quasi_unquote_M(list_expression,
                             Counter,
                             Nesting_level);
    if (car_expression == Pool_Keyword_Symbol(Pool_Quasisymbol))
      return quasi_quasisymbol_M(list_expression,
                                 Counter,
                                 Nesting_level);
    if (car_expression == Pool_Keyword_Symbol(Pool_Unquote_Splicing))
      switch (Splicing_status)
        { case No_splicing:
            return keyword_error(ULS_error,
                                 Pool_Quasiquote);
          case List_splicing:
            return quasi_unquote_list_splicing_M(list_expression,
                                                 Counter,
                                                 Nesting_level);
          case Vector_splicing:
            return quasi_unquote_vector_splicing_M(list_expression,
                                                   Counter,
                                                   Nesting_level); }
    return quasi_pair_M(Pair,
                        Counter,
                        Nesting_level); }

/*------------------------------------quasi-vector -------------------------------------*/

static VEC_type quasi_vector_M(VEC_type Vector,
                               UNS_type Counter,
                               UNS_type Nesting_level)
  { EXP_type expression;
    UNS_type index,
             size;
    VEC_type quasiquoted_vector,
             vector;
    size = size_VEC(Vector);
    if (size == 0)
      return Unique_Empty_Vector_P;
    Stack_Push_M(Vector);
    quasiquoted_vector = make_VEC_M(size);
    vector = Stack_Swap(quasiquoted_vector);
    for (index = 1;
         index <= size;
         index += 1)
      { Stack_Push_M(vector);
        expression = vector[index];
        expression = quasi_expression_M(expression,
                                        Vector_splicing,
                                        Counter,
                                        Nesting_level);
        vector = Stack_Pop();
        quasiquoted_vector = Stack_Peek();
        quasiquoted_vector[index] = expression; }
    Stack_Zap();
    return quasiquoted_vector; }

/*---------------------------------- quasi-expression ----------------------------------*/

static EXP_type quasi_expression_M(EXP_type Expression,
                                   SPL_type Splicing_status,
                                   UNS_type Counter,
                                   UNS_type Nesting_level)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          return quasi_form_M(Expression,
                              Splicing_status,
                              Counter,
                              Nesting_level);
        case VEC_tag:
          return quasi_vector_M(Expression,
                                Counter,
                                Nesting_level);
        case CHA_tag:
        case FLS_tag:
        case NUL_tag:
        case REA_tag:
        case SMI_tag:
        case STR_tag:
        case SYM_tag:
        case TRU_tag:
          return quasi_literal(Expression);
        default:
          return keyword_error(IXT_error,
                               Pool_Quasiquote); }}

/*--------------------------------------------------------------------------------------*/

static EXP_type quasiquote_M(VOI_type)
  { EXP_type expression;
    QQU_type quasiquote;
    UNS_type counter;
    Expression_P = single(List_Expression_P,
                          Pool_Quasiquote);
    counter = Main_Count();
    quasiquote = make_QQU_M();
    expression = quasi_expression_M(Expression_P,
                                    No_splicing,
                                    counter,
                                    1);
    quasiquote->exp = expression;
    return quasiquote; }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- quasisymbol -----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type keyword_error(UNS_type,
                              KEY_type);
static EXP_type        single(EXP_type,
                              BLN_type);

/*--------------------------------------------------------------------------------------*/

static QSY_type quasisymbol_M(VOI_type)
  { EXP_type symbol_expression;
    QSY_type quasisymbol;
    SMI_type counter_smallint;
    UNS_type counter;
    symbol_expression = single(List_Expression_P,
                               Pool_Quasisymbol);
    if (!is_SYM(symbol_expression))
      return keyword_error(NSY_error,
                           Pool_Quasisymbol);
    Symbol_P = symbol_expression;
    counter = Main_Count();
    counter_smallint = make_SMU(counter);
    quasisymbol = make_QSY_M();
    quasisymbol->sym = Symbol_P;
    quasisymbol->ctr = counter_smallint;
    return quasisymbol; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- quote ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type single(EXP_type,
                       BLN_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type quote_M(EXP_type List_expression)
  { QUO_type quote;
    Expression_P = single(List_Expression_P,
                        Pool_Quote);
    quote = make_QUO_M();
    quote->exp = Expression_P;
    return quote; }
    
/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- set! ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type  expression_M(EXP_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair3(EXP_type,
                              EXp_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type set_M(VOI_type)
  { EXP_type symbol_expression;
    SMI_type offset_smallint,
             scope_smallint;
    STG_type set_global;
    STL_type set_local;
    UNS_type offset,
             scope,
             size;
    size = list(List_Expression_P,
                Pool_Set);
    if (size != 2)
      return keyword_error(EX2_error,
                           Pool_Set);
    unpair3(List_Expression_P,
            &symbol_expression,
            &Expression_P,
            VOID);
    if (!is_SYM(symbol_expression))
      return keyword_error(NSY_error,
                           Pool_Set);
    if (Dictionary_Lexical_Address_M(symbol_expression,
                                    &scope,
                                    &offset))
       symbol_error(VNF_error,
                    symbol_expression);
    offset_smallint = make_SMU(offset);
    Expression_P = expression_M(Expression_P,
                                FALSE);
    if (scope == 0)
      { set_local = make_STL_M();
        set_local->ofs = offset_smallint;
        set_local->exp = Expression_P;
        return set_local; }
    scope_smallint = make_SMU(scope);
    set_global = make_STG_M();
    set_global->scp = scope_smallint;
    set_global->ofs = offset_smallint;
    set_global->exp = Expression_P;
    return set_global; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- while ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type      inline_M(EXP_type,
                              BLN_type);
static EXP_type inline_body_M(EXP_type,
                              KEY_type,
                              BLN_type);
static EXP_type keyword_error(UNS_type,
                              KEY_type);
static UNS_type          list(EXP_type,
                              KEY_type);
static VOI_type       unpair2(EXP_type,
                              EXp_type,
                              EXp_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type while_M(VOI_type)
  { EXP_type expression,
             predicate_expression;
    UNS_type size;
    WHI_type while_;
    size = list(List_Expression_P,
                Pool_While);
    if (size < 2)
      return keyword_error(TFX_error,
                           Pool_While);
    unpair2(List_Expression_P,
            &Expression_P,
            &expression);
    Stack_Push_M(expression);
    predicate_expression = inline_M(Expression_P,
                                    FALSE);
    expression = Stack_Swap(predicate_expression);
    Expression_P = inline_body_M(expression,
                                 Pool_While,
                                 FALSE);
    while_= make_WHI_M();
    while_->prd = Stack_Pop();
    while_->exp = Expression_P;
    return while_; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- form ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static AND_type               and_M(BLN_type);
static EXP_type       application_M(BLN_type);
static BEG_type             begin_M(BLN_type);
static EXP_type              cond_M(BLN_type);
static CNS_type       cons_stream_M(VOI_type);
static EXP_type            define_M(VOI_type);
static DEL_type             delay_M(VOI_type);
static EXP_type              eval_M(BLN_type);
static EXP_type                if_M(BLN_type);
static EXP_type       keyword_error(UNS_type,
                                    KEY_type);
static EXP_type            lambda_M(VOI_type);
static LET_type               let_M(BLN_type);
static EXP_type              load_M(BLN_type);
static MAC_type             macro_M(VOI_type);
static EXP_type macro_application_M(EXP_type,
                                    BLN_type);
static ORR_type                or_M(EXP_type,
                                    BLN_type);
static EXP_type        quasiquote_M(VOI_type);
static QSY_type       quasisymbol_M(VOI_type);
static EXP_type             quote_M(EXP_type);
static EXP_type               set_M(VOI_type);
static VOI_type             unpair2(EXP_type,
                                    EXp_type,
                                    EXp_type);
static EXP_type             while_M(VOI_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type form_M(PAI_type Form_pair,
                       BLN_type Boolean)
  { EXP_type macro_expression;
    unpair2(Form_pair,
            &Expression_P,
            &List_Expression_P);
    if (is_SYM(Expression_P))
      { if (Expression_P == Pool_Keyword_Symbol(Pool_And))
          return and_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Begin))
          return begin_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Cond))
          return cond_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Cons_Stream))
          return cons_stream_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Define))
          return define_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Delay))
          return delay_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Eval))
          return eval_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Else))
          return keyword_error(IUK_error,
                               Pool_Else);
        if (Expression_P == Pool_Keyword_Symbol(Pool_If))
          return if_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Lambda))
          return lambda_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Let))
          return let_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Load))
          return load_M(Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Macro))
          return macro_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Or))
          return or_M(List_Expression_P,
                      Boolean);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Quasiquote))
          return quasiquote_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Quasisymbol))
          return quasisymbol_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Quote))
          return quote_M(List_Expression_P);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Set))
          return set_M();
        if (Expression_P == Pool_Keyword_Symbol(Pool_Unquote))
          return keyword_error(UQQ_error,
                               Pool_Unquote);
        if (Expression_P == Pool_Keyword_Symbol(Pool_Unquote_Splicing))
          return keyword_error(UQQ_error,
                               Pool_Unquote_Splicing);
        if (Expression_P == Pool_Keyword_Symbol(Pool_While))
          return while_M();
        macro_expression = Dictionary_Get_Macro_Expression(Expression_P);
        if (!is_UDF(macro_expression))
          return macro_application_M(macro_expression,
                                     Boolean); }
    return application_M(Boolean); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- literal --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type literal(EXP_type Literal_expression)
  { return Literal_expression; }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- variable --------------------------------------*/
/*--------------------------------------------------------------------------------------*/


static EXP_type macro_value_M(MVL_type);
static EXP_type  symbol_error(UNS_type,
                              SYM_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type variable_M(SYM_type Variable_symbol)
  { EXP_type macro_expression;
    GLB_type global;
    LCL_type local;
    UNS_type offset,
             scope;
    if (Pool_Validate_Symbol(Variable_symbol) == TRUE)
      symbol_error(IUK_error,
                   Variable_symbol);
    macro_expression = Dictionary_Get_Macro_Expression(Variable_symbol);
    if (!is_UDF(macro_expression))
      { if (!is_MVL(macro_expression))
          return symbol_error(IMV_error,
                              Variable_symbol);
        return macro_value_M(macro_expression); }
    if (Dictionary_Lexical_Address_M(Variable_symbol,
                                     &scope,
                                     &offset))
      symbol_error(VNF_error,
                   Variable_symbol);
    if (scope == 0)
      { local = make_LCL_M();
        local->ofs = offset;
        return local; }
    global = make_GLB_M(); 
    global->scp = scope;
    global->ofs = offset;
    return global; }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- expression ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type     form_M(PAI_type,
                           BLN_type);
static EXP_type    literal(EXP_type);
static EXP_type variable_M(SYM_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type expression_M(EXP_type Expression,
                             BLN_type Boolean)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          return form_M(Expression,
                        Boolean);
        case SYM_tag:
          return variable_M(Expression);
        case CHA_tag:
        case FLS_tag:
        case NST_tag:
        case NUL_tag:
        case REA_tag:
        case SMI_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
          return literal(Expression);
        default:
          return error(IXT_error,
                       Compile_text); }}

/*---------------------------------- public functions ----------------------------------*/

EXP_type Compile_M(EXP_type Expression)
  { return expression_M(Expression,
                        FALSE); }

EXP_type Compile_T_M(EXP_type Expression)
  { return expression_M(Expression,
                        TRUE); }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Compile_Initialize_M(VOI_type)
  { Dictionary_Macro_Initialize_M();
    REGISTER(Body_Expression_P);
    REGISTER(Car_Expression_P);
    REGISTER(Cdr_Expression_P);
    REGISTER(Dictionary_P);
    REGISTER(Environment_P);
    REGISTER(Expression_P);
    REGISTER(Frame_P);
    REGISTER(Input_Port_P);
    REGISTER(List_Expression_P);
    REGISTER(Macro_Value_P);
    REGISTER(Pair_P);
    REGISTER(Predicate_Expression_P);
    REGISTER(Procedure_P);
    REGISTER(Procedure_Vararg_P);
    REGISTER(Sequence_P);
    REGISTER(String_P);
    REGISTER(Symbol_P); }

