                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    read                     */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipGrammar.h"

#include "SlipPool.h"
#include "SlipScan.h"
#include "SlipStack.h"

#include "SlipRead.h"

/*----------------------------------- s-expressions ------------------------------------*/
/*                                                                                      */
/* [s-expression]         ::= [simple expression] | [composite expression]              */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* [simple expression]    ::= [atom] | [quasisymbol] | [symbol]                         */
/*                                                                                      */
/* [composite expression] ::= [list] | [quotation] | [unquotation] | [vector]           */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* [atom]             ::= [character] | [false] | [null] | [number] | [string] | [true] */
/*                                                                                      */
/* [list]             ::= [proper-list] | [improper-list]                               */
/*                                                                                      */
/* [quotation]        ::= [quote] | [quasiquote]                                        */
/*                                                                                      */
/* [quasisymbol]      ::= `@ {symbol}                                                   */
/*                                                                                      */
/* [symbol]           ::= {symbol}                                                      */
/*                                                                                      */
/* [unquotation]      ::= [unquote] | [unquote-splicing]                                */
/*                                                                                      */
/* [vector]           ::= #( [s-expression]* )                                          */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* [character]        ::= #\ {any character} | #\ space | #\ newline                    */
/*                                                                                      */
/* [false]            ::= #f                                                            */
/*                                                                                      */
/* [improper-list]    ::= ( [s-expression]+ . [s-expression] )                          */
/*                                                                                      */
/* [null]             ::= ()                                                            */
/*                                                                                      */
/* [number]           ::= {integer} | {float}                                           */
/*                                                                                      */
/* [proper-list]      ::= ( [s-expression]+ )                                           */
/*                                                                                      */
/* [quote]            ::= ' [s-expression]                                              */
/*                                                                                      */
/* [quasiquote]       ::= ` [s-expression]                                              */
/*                                                                                      */
/* [string]           ::= {string}                                                      */
/*                                                                                      */
/* [true]             ::= #t                                                            */
/*                                                                                      */
/* [unquote]          ::= , [s-expression]                                              */
/*                                                                                      */
/* [unquote-splicing] ::= ,@ [s-expression]                                             */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/

/*---------------------------------- private constants ---------------------------------*/

static const TXT_type Newline_text = "newline";
static const TXT_type Read_text    = "Read";
static const TXT_type Space_text   = "space";

/*---------------------------------- private variables ---------------------------------*/

static TKN_type Token;

/*---------------------------------- protected variables --------------------------------*/

static EXP_type Expression_P;
static PAI_type Pair_P;

/*---------------------------------- private prototypes --------------------------------*/

static EXP_type read_M(VOI_type);

/*---------------------------------- private functions ---------------------------------*/

static EXP_type error(UNS_type Error)
  { Main_Error(Error,
               Read_text);
    return Unique_Undefined; }

static VOI_type next_token(VOI_type)
  { Token = Scan_Token(); }

static BLN_type reserved_name(TXT_type Text)
  { return !C_STRCMP(Scan_Text,
                     Text); }

/*------------------------------------- generic quote ----------------------------------*/

static PAI_type quotation_M(KEY_type Key)
  { EXP_type expression;
    PAI_type pair;
    next_token();
    expression = read_M();
    pair = Pool_Quotation_M(Key,
                            expression);
    return pair; }

/*--------------------------------------- character ------------------------------------*/

static CHA_type character(VOI_type)
  { CHA_type character;
    CHR_type char_;
    if (reserved_name(Space_text))
      character = Unique_Space_Character;
    else if (reserved_name(Newline_text))
      character =  Unique_Newline_Character;
    else if (C_STRLEN(Scan_Text) != 1)
      return error(ICL_error);
    else
      { char_ = Scan_Text[0];
        character = make_CHA(char_); }
    next_token();
    return character; }

/*-------------------------------------- end of input ----------------------------------*/

static EXP_type end_of_input(VOI_type)
  { return error(PEP_error); }

/*----------------------------------------- false --------------------------------------*/

static FLS_type false(VOI_type)
  { next_token();
    return Unique_False; }

/*----------------------------------------- list ---------------------------------------*/

static EXP_type list_M(VOI_type)
  { EXP_type expression;
    PAI_type pair;
    UNS_type size;
    next_token();
    for (size = 0;
         ;
         size += 1)
      if (Token == RPR_token)
        { Expression_P = Unique_Null;
          break; }
      else
        { expression = read_M();
          Stack_Push_M(expression);
          if (Token == PER_token)
            { next_token();
              Expression_P = read_M();
              size += 1;
              if (Token != RPR_token)
                error(MRP_error);
              break; }}
    next_token();
    for (;
         size > 0;
         size -= 1)
      { pair = make_PAI_M();
        pair->car = Stack_Pop();
        pair->cdr = Expression_P;
        Expression_P = pair; }
    return Expression_P; }

/*--------------------------------------- integer --------------------------------------*/

static SMI_type integer(VOI_type)
  { SGN_type integer;
    LNG_type long_;
    long_ = C_STRTOL(Scan_Text, VOID, 10);
    if (C_ERANGE)
      error(NTL_error);
    if (C_LLABS(long_) > Maximum_Smallint_Size)
      return error(NTL_error);
    integer = (SGN_type)long_;
    next_token();
    return make_SMI(integer); }

/*----------------------------------- right parenthesis --------------------------------*/

static EXP_type right_parenthesis(VOI_type)
  { return error(IRP_error); }

/*---------------------------------------- period --------------------------------------*/

static EXP_type period(VOI_type)
  { return error(IUP_error); }

/*-------------------------------------- quasiquote ------------------------------------*/

static PAI_type quasiquote_M(VOI_type)
  { return quotation_M(Pool_Quasiquote); }

/*-------------------------------------- quasisymbol -----------------------------------*/

static PAI_type quasisymbol_M(VOI_type)
  { return quotation_M(Pool_Quasisymbol); }

/*----------------------------------------- quote --------------------------------------*/

static PAI_type quote_M(VOI_type)
  { return quotation_M(Pool_Quote); }

/*----------------------------------------- real ---------------------------------------*/

static REA_type real_M(VOI_type)
  { FLO_type float_;
    REA_type real;
    float_ = C_STRTOD(Scan_Text, VOID);
    if (C_ERANGE)
      return error(NTL_error);
    next_token();
    real = make_REA_M();
    real->flo = float_;
    return real; }

/*---------------------------------------- string --------------------------------------*/

static STR_type string_M(VOI_type)
  { STR_type string;
    UNS_type length;
    length = C_STRLEN(Scan_Text);
    if (length == 0)
      string = Unique_Empty_String_P;
    else
      { string = make_STR_M(length);
        C_STRCPY(string->txt, Scan_Text); }
    next_token();
    return string; }

/*----------------------------------------- symbol -------------------------------------*/

static SYM_type symbol_M(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_Text_M(Scan_Text);
    next_token();
    return symbol; }

/*------------------------------------------ true --------------------------------------*/

static TRU_type true(VOI_type)
  { next_token();
    return Unique_True; }

/*---------------------------------------- unquote -------------------------------------*/

static PAI_type unquote_M(VOI_type)
  { return quotation_M(Pool_Unquote); }

/*------------------------------------ unquote-splicing --------------------------------*/

static PAI_type unquote_splice_M(VOI_type)
  { return quotation_M(Pool_Unquote_Splicing); }

/*---------------------------------------- vector --------------------------------------*/

static VEC_type vector_M(VOI_type)
  { EXP_type expression;
    UNS_type size;
    VEC_type vector;
    next_token();
    for (size = 0;
         Token != RPR_token;
         size += 1)
      { expression = read_M();
        Stack_Push_M(expression); }
    next_token();
    if (size == 0)
      return Unique_Empty_Vector_P;
    vector = make_VEC_Z(size);
    for (;
         size > 0;
         size -= 1)
      { expression = Stack_Pop();
        vector[size] = expression; }
    return vector; }

/*-------------------------------------- expression ------------------------------------*/

static EXP_type read_M(VOI_type)
  { switch (Token)
      { case CHA_token:
          return character();
        case EOI_token:
          return end_of_input();
        case FLS_token:
          return false();
        case INT_token:
          return integer();
        case LPR_token:
          return list_M();
        case PER_token:
          return period();
        case QQU_token:
          return quasiquote_M();
        case QSY_token:
          return quasisymbol_M();
        case QUO_token:
          return quote_M();
        case RPR_token:
          return right_parenthesis();
        case REA_token:
          return real_M();
        case STR_token:
          return string_M();
        case SYM_token:
          return symbol_M();
        case TRU_token:
          return true();
        case UNQ_token:
          return unquote_M();
        case UQS_token:
          return unquote_splice_M();
        case VEC_token:
          return vector_M();
        default:
          return error(ITK_error); }}

/*---------------------------------- public functions ----------------------------------*/

EXP_type Read_Port_M(IPT_type Input_port)
  { ADR_type address;
    EXP_type expression;
    address = Input_port->adr;
    Token = Scan_Input_Port(address);
    expression = read_M();
    if (Token != EOI_token)
      error(XIT_error);
    return expression; }

EXP_type Read_String_M(TXT_type Input_text)
  { EXP_type expression;
    Stack_Reset();
    Token = Scan_String(Input_text);
    expression = read_M();
    if (Token != EOI_token)
      error(XIT_error);
    return expression; }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Read_Initialize_M(VOI_type)
  { Scan_Initialize();
    REGISTER(Expression_P);
    REGISTER(Pair_P); }

